(declare (block) (standard-bindings) (extended-bindings))
(begin
  (define-macro (define-guard guard defn)
    (if (eval `(cond-expand (,guard #t) (else #f)))
        '(begin)
        (begin (eval `(define-cond-expand-feature ,guard)) defn)))
  (define-macro (define-c-lambda id args ret #!optional (name #f))
    (let ((name (or name (symbol->string id))))
      `(define ,id (c-lambda ,args ,ret ,name))))
  (define-macro (define-const symbol)
    (let* ((str (symbol->string symbol))
           (ref (string-append "___return (" str ");")))
      `(define ,symbol ((c-lambda () int ,ref)))))
  (define-macro (define-const* symbol #!optional (ccond #f))
    (let* ((str (symbol->string symbol))
           (code (string-append
                  "#if "
                  (or ccond (string-append "defined(" str ")"))
                  "\n"
                  "___return (___FIX ("
                  str
                  "));\n"
                  "#else \n"
                  "___return (___FAL);\n"
                  "#endif")))
      `(define ,symbol ((c-lambda () scheme-object ,code)))))
  (define-macro (define-with-errno symbol ffi-symbol args)
    `(define (,symbol ,@args)
       (declare (not interrupts-enabled))
       (let ((r (,ffi-symbol ,@args)))
         (if (##fx< r 0) (##fx- (##c-code "___RESULT = ___FIX (errno);")) r))))
  (define-macro (define-c-struct
                 struct
                 #!optional
                 (members '())
                 release-function)
    (let* ((struct-str (symbol->string struct))
           (struct-ptr (string->symbol (string-append struct-str "*")))
           (shallow-ptr
            (string->symbol (string-append struct-str "-shallow-ptr*")))
           (borrowed-ptr
            (string->symbol (string-append struct-str "-borrowed-ptr*")))
           (string-types
            '(char-string
              nonull-char-string
              UTF-8-string
              nonnull-UTF-8-string
              UTF-16-string
              nonnull-UTF16-string))
           (string-compat-required?
            (let loop ((m members))
              (cond ((null? m) #f)
                    ((member (cdr (car m)) string-types) #t)
                    (else (loop (cdr m))))))
           (string-setter-body
            (lambda (member-name)
              (let ((m (string-append "___arg1->" member-name)))
                (string-append
                 "if("
                 m
                 " == NULL)"
                 "\n"
                 m
                 "= strdup(___arg2);"
                 "\n"
                 "else if (strcmp("
                 m
                 ", ___arg2) != 0) {"
                 "\n"
                 "free("
                 m
                 ");"
                 "\n"
                 m
                 "= strdup(___arg2);"
                 "\n"
                 "}"
                 "\n"
                 "___return;"
                 "\n"))))
           (default-free-body
            (and string-compat-required?
                 (string-append
                  "___SCMOBJ "
                  struct-str
                  "_ffi_free (void *ptr) {"
                  "\n"
                  "struct "
                  struct-str
                  " *obj = (struct "
                  struct-str
                  "*) ptr;"
                  "\n"
                  (apply string-append
                         (map (lambda (m)
                                (cond ((memq (cdr m) string-types)
                                       (let ((mem-name
                                              (symbol->string (car m))))
                                         (string-append
                                          "if(obj->"
                                          mem-name
                                          ") "
                                          "free(obj->"
                                          mem-name
                                          ");"
                                          "\n")))
                                      (else "")))
                              members))
                  "free(obj);"
                  "\n"
                  "return ___FIX (___NO_ERR);"
                  "\n"
                  "}")))
           (release-function
            (or release-function
                (if string-compat-required?
                    (string-append struct-str "_ffi_free")
                    "ffi_free")))
           (string-compat-types
            (if string-compat-required?
                `((c-declare ,default-free-body)
                  (c-define-type
                   ,shallow-ptr
                   (pointer ,struct (,struct-ptr) "ffi_free")))
                '())))
      `(begin
         (c-define-type ,struct (struct ,struct-str))
         (c-define-type
          ,struct-ptr
          (pointer ,struct (,struct-ptr) ,release-function))
         (c-define-type ,borrowed-ptr (pointer ,struct (,struct-ptr)))
         ,@string-compat-types
         (define ,(string->symbol (string-append struct-str "-ptr?"))
           (lambda (obj)
             (and (foreign? obj) (equal? (foreign-tags obj) '(,struct-ptr)))))
         ,@(apply append
                  (map (lambda (m)
                         (let* ((member-name (symbol->string (car m)))
                                (member-type (cdr m))
                                (getter-name
                                 (string-append struct-str "-" member-name))
                                (setter-body
                                 (cond ((member member-type string-types)
                                        (string-setter-body member-name))
                                       (else
                                        (string-append
                                         "___arg1->"
                                         member-name
                                         " = ___arg2;"
                                         "\n"
                                         "___return;"
                                         "\n")))))
                           `((define ,(string->symbol getter-name)
                               (c-lambda
                                (,struct-ptr)
                                ,member-type
                                ,(string-append
                                  "___return(___arg1->"
                                  member-name
                                  ");")))
                             (define ,(string->symbol
                                       (string-append getter-name "-set!"))
                               (c-lambda
                                (,struct-ptr ,member-type)
                                void
                                ,setter-body)))))
                       members))
         (define ,(string->symbol (string-append "malloc-" struct-str))
           (c-lambda
            ()
            ,struct-ptr
            ,(string-append
              "struct "
              struct-str
              "* var = (struct "
              struct-str
              " *) malloc(sizeof(struct "
              struct-str
              "));"
              "\n"
              "if (var == NULL)"
              "\n"
              "    ___return (NULL);"
              "\n"
              "memset(var, 0, sizeof(struct "
              struct-str
              "));"
              "___return(var);")))
         (define ,(string->symbol (string-append "ptr->" struct-str))
           (c-lambda (,struct-ptr) ,struct "___return(*___arg1);"))
         (define ,(string->symbol
                   (string-append "malloc-" struct-str "-array"))
           (c-lambda
            (unsigned-int32)
            ,(if string-compat-required? shallow-ptr struct-ptr)
            ,(string-append
              "struct "
              struct-str
              " *arr_var=(struct "
              struct-str
              " *) malloc(___arg1*sizeof(struct "
              struct-str
              "));"
              "\n"
              "if (arr_var == NULL)"
              "\n"
              "    ___return (NULL);"
              "\n"
              "memset(arr_var, 0, ___arg1*sizeof(struct "
              struct-str
              "));"
              "\n"
              "___return(arr_var);")))
         (define ,(string->symbol (string-append struct-str "-array-ref"))
           (c-lambda
            (,struct-ptr unsigned-int32)
            ,borrowed-ptr
            "___return (___arg1 + ___arg2);"))
         (define ,(string->symbol (string-append struct-str "-array-set!"))
           (c-lambda
            (,struct-ptr unsigned-int32 ,struct-ptr)
            void
            "*(___arg1 + ___arg2) = *___arg3; ___return;")))))
  (c-declare "#include <stdlib.h>")
  (c-declare "#include <string.h>")
  (c-declare "#include <errno.h>")
  (c-declare "static ___SCMOBJ ffi_free (void *ptr);")
  (c-declare
   "#ifndef ___HAVE_FFI_U8VECTOR\n#define ___HAVE_FFI_U8VECTOR\n#define U8_DATA(obj) ___CAST (___U8*, ___BODY_AS (obj, ___tSUBTYPED))\n#define U8_LEN(obj) ___HD_BYTES (___HEADER (obj))\n#endif")
  (namespace
   ("std/crypto/libcrypto#"
    EVP_MD_CTX_set_pkey_ctx
    EVP_DigestVerify
    EVP_DigestSign
    EVP_PKEY_get_raw_public_key
    EVP_PKEY_get_raw_private_key
    EVP_PKEY_new_raw_public_key
    EVP_PKEY_new_raw_private_key
    EVP_PKEY_keygen
    EVP_PKEY_keygen_init
    EVP_PKEY_CTX_dup
    EVP_PKEY_CTX_new_id
    EVP_PKEY_CTX_new
    EVP_PKEY_CTX*
    EVP_PKEY_CTX
    EVP_PKEY*
    EVP_PKEY
    EVP_PKEY_ED448
    EVP_PKEY_X448
    EVP_PKEY_ED25519
    EVP_PKEY_X25519
    EVP_PKEY_SIPHASH
    EVP_PKEY_POLY1305
    EVP_PKEY_HKDF
    EVP_PKEY_TLS1_PRF
    EVP_PKEY_SCRYPT
    EVP_PKEY_CMAC
    EVP_PKEY_HMAC
    EVP_PKEY_SM2
    EVP_PKEY_EC
    EVP_PKEY_DHX
    EVP_PKEY_DH
    EVP_PKEY_DSA4
    EVP_PKEY_DSA3
    EVP_PKEY_DSA2
    EVP_PKEY_DSA1
    EVP_PKEY_DSA
    EVP_PKEY_RSA_PSS
    EVP_PKEY_RSA2
    EVP_PKEY_RSA
    EVP_PKEY_NONE
    DH_compute_key
    DH_size
    DH_generate_key
    DH_get_2048_256
    DH_get_2048_224
    DH_get_1024_160
    DH_new
    DH_pub_key
    DH?
    DH*
    BN_bn2bin
    BN_bin2bn
    BN_num_bytes
    BN?
    HMAC
    HMAC_Final
    HMAC_Update
    HMAC_Init
    HMAC_CTX_create
    HMAC_CTX?
    EVP_get_cipherbynid
    EVP_get_cipherbyname
    EVP_CIPHER_CTX_copy
    EVP_CIPHER_CTX_iv_length
    EVP_CIPHER_CTX_key_length
    EVP_CIPHER_CTX_block_size
    EVP_CIPHER_CTX_nid
    EVP_CIPHER_CTX_cipher
    EVP_CIPHER_name
    EVP_CIPHER_iv_length
    EVP_CIPHER_key_length
    EVP_CIPHER_block_size
    EVP_CIPHER_nid
    EVP_bf_ofb
    EVP_bf_cfb
    EVP_bf_cbc
    EVP_bf_ecb
    EVP_cast5_ofb
    EVP_cast5_cfb
    EVP_cast5_cbc
    EVP_cast5_ecb
    EVP_camellia_256_ofb
    EVP_camellia_256_cfb
    EVP_camellia_256_cbc
    EVP_camellia_256_ecb
    EVP_camellia_192_ofb
    EVP_camellia_192_cfb
    EVP_camellia_192_cbc
    EVP_camellia_192_ecb
    EVP_camellia_128_ofb
    EVP_camellia_128_cfb
    EVP_camellia_128_cbc
    EVP_camellia_128_ecb
    EVP_aes_256_xts
    EVP_aes_256_gcm
    EVP_aes_256_ccm
    EVP_aes_256_ctr
    EVP_aes_256_ofb
    EVP_aes_256_cfb
    EVP_aes_256_cbc
    EVP_aes_256_ecb
    EVP_aes_192_gcm
    EVP_aes_192_ccm
    EVP_aes_192_ctr
    EVP_aes_192_ofb
    EVP_aes_192_cfb
    EVP_aes_192_cbc
    EVP_aes_192_ecb
    EVP_aes_128_xts
    EVP_aes_128_gcm
    EVP_aes_128_ccm
    EVP_aes_128_ctr
    EVP_aes_128_ofb
    EVP_aes_128_cfb
    EVP_aes_128_cbc
    EVP_aes_128_ecb
    EVP_rc4
    EVP_DecryptFinal
    EVP_DecryptUpdate
    EVP_DecryptInit
    EVP_EncryptFinal
    EVP_EncryptUpdate
    EVP_EncryptInit
    EVP_CIPHER_CTX_create
    EVP_CIPHER_CTX?
    EVP_CIPHER?
    EVP_get_digestbynid
    EVP_get_digestbyname
    EVP_MD_CTX_block_size
    EVP_MD_CTX_size
    EVP_MD_CTX_type
    EVP_MD_CTX_md
    EVP_MD_name
    EVP_MD_block_size
    EVP_MD_size
    EVP_MD_pkey_type
    EVP_MD_type
    EVP_keccak256
    EVP_shake256
    EVP_shake128
    EVP_sha3_512
    EVP_sha3_384
    EVP_sha3_256
    EVP_sha3_224
    EVP_blake2s256
    EVP_blake2b512
    EVP_whirlpool
    EVP_ripemd160
    EVP_sha512
    EVP_sha384
    EVP_sha256
    EVP_sha224
    EVP_sha1
    EVP_md5
    EVP_MD_CTX_copy
    EVP_DigestFinal
    EVP_DigestUpdate
    EVP_DigestInit
    EVP_MD_CTX_create
    EVP_MD_CTX?
    EVP_MD?
    ERR_reason_error_string
    ERR_func_error_string
    ERR_lib_error_string
    ERR_peek_last_error
    ERR_get_error))
  (declare (not safe))
  (c-declare
   "#include <openssl/evp.h>\n#include <openssl/err.h>\n#include <openssl/dh.h>\n#include <openssl/bn.h>\n#include <openssl/hmac.h>")
  (c-initialize
   "ERR_load_crypto_strings(); /* Load the human readable error strings for libcrypto */\nOpenSSL_add_all_algorithms(); /* Load all digest and cipher algorithms, + engines on BSD */")
  (c-declare
   "#ifndef ___HAVE_FFI_U8VECTOR\n#define ___HAVE_FFI_U8VECTOR\n#define U8_DATA(obj) ___CAST (___U8*, ___BODY_AS (obj, ___tSUBTYPED))\n#define U8_LEN(obj) ___HD_BYTES (___HEADER (obj))\n#endif")
  (define-macro (define-c-lambda/const-pointer
                 id
                 args
                 ret
                 #!optional
                 (ccond #f)
                 (iffalse "NULL"))
    (let* ((result-type
            (if (eq? ret 'char-string) "char*" (##symbol->string ret)))
           (c-args (let lp ((rest args) (n 1) (c-args ""))
                     (if (##pair? rest)
                         (let* ((next (##string-append
                                       "___arg"
                                       (##number->string n)))
                                (c-args (if (##fx> n 1)
                                            (##string-append c-args "," next)
                                            next)))
                           (lp (##cdr rest) (##fx+ n 1) c-args))
                         c-args)))
           (expr (##string-append
                  "___return (("
                  result-type
                  ")"
                  (##symbol->string id)
                  "("
                  c-args
                  "));")))
      `(define ,id
         (c-lambda
          ,args
          ,ret
          ,(if ccond
               (##string-append
                "#if "
                ccond
                "\n"
                expr
                "\n#else\n___return("
                iffalse
                ");\n#endif\n")
               expr)))))
  (define-macro (define-c-type-predicate pred tag)
    `(define (,pred x) (and (##foreign? x) (##memq ',tag (foreign-tags x)))))
  (define-c-lambda ERR_get_error () unsigned-long)
  (define-c-lambda ERR_peek_last_error () unsigned-long)
  (define-c-lambda/const-pointer
   ERR_lib_error_string
   (unsigned-long)
   char-string)
  (define-c-lambda/const-pointer
   ERR_func_error_string
   (unsigned-long)
   char-string)
  (define-c-lambda/const-pointer
   ERR_reason_error_string
   (unsigned-long)
   char-string)
  (c-define-type ENGINE "ENGINE")
  (c-define-type ENGINE* (pointer ENGINE (ENGINE*)))
  (c-declare
   "static ___SCMOBJ ffi_release_EVP_MD_CTX (void *ptr)\n{\n  EVP_MD_CTX_destroy ((EVP_MD_CTX*)ptr); /* NB: renamed _free in 1.1.0 */\n  return ___FIX (___NO_ERR);\n}\n\nstatic int ffi_EVP_DigestInit (EVP_MD_CTX *ctx, EVP_MD *type) {\n  return EVP_DigestInit_ex (ctx, type, NULL);\n}\n\nstatic int ffi_EVP_DigestUpdate (EVP_MD_CTX *ctx, ___SCMOBJ bytes, int start, int end)\n{\n  return EVP_DigestUpdate (ctx, U8_DATA (bytes) + start, end - start);\n}\n\nstatic int ffi_EVP_DigestFinal (EVP_MD_CTX *ctx, ___SCMOBJ bytes)\n{\n  return EVP_DigestFinal_ex (ctx, U8_DATA (bytes), NULL);\n}")
  (c-define-type EVP_MD "EVP_MD")
  (c-define-type EVP_MD* (pointer EVP_MD (EVP_MD*)))
  (c-define-type EVP_MD_CTX "EVP_MD_CTX")
  (c-define-type
   EVP_MD_CTX*
   (pointer EVP_MD_CTX (EVP_MD_CTX*) "ffi_release_EVP_MD_CTX"))
  (define-c-type-predicate EVP_MD? EVP_MD*)
  (define-c-type-predicate EVP_MD_CTX? EVP_MD_CTX*)
  (define-c-lambda EVP_MD_CTX_create () EVP_MD_CTX*)
  (define-c-lambda
   EVP_DigestInit
   (EVP_MD_CTX* EVP_MD*)
   int
   "ffi_EVP_DigestInit")
  (define-c-lambda
   EVP_DigestUpdate
   (EVP_MD_CTX* scheme-object int int)
   int
   "ffi_EVP_DigestUpdate")
  (define-c-lambda
   EVP_DigestFinal
   (EVP_MD_CTX* scheme-object)
   int
   "ffi_EVP_DigestFinal")
  (define-c-lambda
   EVP_MD_CTX_copy
   (EVP_MD_CTX* EVP_MD_CTX*)
   int
   "EVP_MD_CTX_copy_ex")
  (define-c-lambda/const-pointer EVP_md5 () EVP_MD* "!defined(OPENSSL_NO_MD5)")
  (define-c-lambda/const-pointer EVP_sha1 () EVP_MD*)
  (define-c-lambda/const-pointer EVP_sha224 () EVP_MD*)
  (define-c-lambda/const-pointer EVP_sha256 () EVP_MD*)
  (define-c-lambda/const-pointer EVP_sha384 () EVP_MD*)
  (define-c-lambda/const-pointer EVP_sha512 () EVP_MD*)
  (define-c-lambda/const-pointer
   EVP_ripemd160
   ()
   EVP_MD*
   "!defined(OPENSSL_NO_RMD160)")
  (define-c-lambda/const-pointer
   EVP_whirlpool
   ()
   EVP_MD*
   "!defined(OPENSSL_NO_WHIRLPOOL)")
  (define-c-lambda/const-pointer
   EVP_blake2b512
   ()
   EVP_MD*
   "(OPENSSL_VERSION_NUMBER >= 0x10100000L) && !defined(OPENSSL_NO_BLAKE2)")
  (define-c-lambda/const-pointer
   EVP_blake2s256
   ()
   EVP_MD*
   "(OPENSSL_VERSION_NUMBER >= 0x10100000L) && !defined(OPENSSL_NO_BLAKE2)")
  (define-c-lambda/const-pointer
   EVP_sha3_224
   ()
   EVP_MD*
   "OPENSSL_VERSION_NUMBER >= 0x10101000L")
  (define-c-lambda/const-pointer
   EVP_sha3_256
   ()
   EVP_MD*
   "OPENSSL_VERSION_NUMBER >= 0x10101000L")
  (define-c-lambda/const-pointer
   EVP_sha3_384
   ()
   EVP_MD*
   "OPENSSL_VERSION_NUMBER >= 0x10101000L")
  (define-c-lambda/const-pointer
   EVP_sha3_512
   ()
   EVP_MD*
   "OPENSSL_VERSION_NUMBER >= 0x10101000L")
  (define-c-lambda/const-pointer
   EVP_shake128
   ()
   EVP_MD*
   "OPENSSL_VERSION_NUMBER >= 0x10101000L")
  (define-c-lambda/const-pointer
   EVP_shake256
   ()
   EVP_MD*
   "OPENSSL_VERSION_NUMBER >= 0x10101000L")
  (define-c-lambda/const-pointer EVP_keccak256 () EVP_MD* "0")
  (define-c-lambda EVP_MD_type (EVP_MD*) int)
  (define-c-lambda EVP_MD_pkey_type (EVP_MD*) int)
  (define-c-lambda EVP_MD_size (EVP_MD*) int)
  (define-c-lambda EVP_MD_block_size (EVP_MD*) int)
  (define-c-lambda/const-pointer EVP_MD_name (EVP_MD*) char-string)
  (define-c-lambda/const-pointer EVP_MD_CTX_md (EVP_MD_CTX*) EVP_MD*)
  (define-c-lambda EVP_MD_CTX_type (EVP_MD_CTX*) int)
  (define-c-lambda EVP_MD_CTX_size (EVP_MD_CTX*) int)
  (define-c-lambda EVP_MD_CTX_block_size (EVP_MD_CTX*) int)
  (define-c-lambda/const-pointer EVP_get_digestbyname (char-string) EVP_MD*)
  (define-c-lambda/const-pointer EVP_get_digestbynid (int) EVP_MD*)
  (c-declare
   "/* like EVP_MD_CTX_create, no EVP_CIPHER_CTX_create available */\nstatic EVP_CIPHER_CTX *ffi_create_EVP_CIPHER_CTX ()\n{\n  EVP_CIPHER_CTX *ctx = EVP_CIPHER_CTX_new ();\n  if (ctx) {\n    EVP_CIPHER_CTX_init (ctx);\n  }\n  return ctx;\n}\n\n/* like EVP_MD_CTX_destroy, no EVP_CIPHER_CTX_destroy available */\nstatic ___SCMOBJ ffi_release_EVP_CIPHER_CTX (void *ptr)\n{\n  EVP_CIPHER_CTX_cleanup ((EVP_CIPHER_CTX*) ptr);\n  EVP_CIPHER_CTX_free ((EVP_CIPHER_CTX*)ptr);\n  return ___FIX (___NO_ERR);\n}\n\nstatic int ffi_EVP_EncryptInit (EVP_CIPHER_CTX *ctx, EVP_CIPHER *type,\n                                ___SCMOBJ key, ___SCMOBJ iv)\n{\n  return EVP_EncryptInit_ex (ctx, type, NULL, U8_DATA (key), U8_DATA (iv));\n}\n\nstatic int ffi_EVP_EncryptUpdate (EVP_CIPHER_CTX *ctx,\n                                  ___SCMOBJ out, int out_start,\n                                  ___SCMOBJ in, int start, int end)\n{\n  int r, olen;\n  r = EVP_EncryptUpdate (ctx, U8_DATA (out) + out_start, &olen,\n                         U8_DATA (in) + start, end - start);\n  if (r) {\n    return olen;\n  } else {\n    return -1;\n  }\n}\n\nstatic int ffi_EVP_EncryptFinal (EVP_CIPHER_CTX *ctx, ___SCMOBJ out, int start)\n{\n  int r, olen;\n  r = EVP_EncryptFinal_ex (ctx, U8_DATA (out) + start, &olen);\n  if (r) {\n    return olen;\n  } else {\n    return -1;\n  }\n}\n\nstatic int ffi_EVP_DecryptInit (EVP_CIPHER_CTX *ctx, EVP_CIPHER *type,\n                                ___SCMOBJ key, ___SCMOBJ iv)\n{\n  return EVP_DecryptInit_ex (ctx, type, NULL, U8_DATA (key), U8_DATA (iv));\n}\n\nstatic int ffi_EVP_DecryptUpdate (EVP_CIPHER_CTX *ctx,\n                                  ___SCMOBJ out, int out_start,\n                                  ___SCMOBJ in, int start, int end)\n{\n  int r, olen;\n  r = EVP_DecryptUpdate (ctx, U8_DATA (out) + out_start, &olen,\n                         U8_DATA (in) + start, end - start);\n  if (r) {\n    return olen;\n  } else {\n    return -1;\n  }\n}\n\nstatic int ffi_EVP_DecryptFinal (EVP_CIPHER_CTX *ctx, ___SCMOBJ out, int start)\n{\n  int r, olen;\n  r = EVP_DecryptFinal_ex (ctx, U8_DATA (out) + start, &olen);\n  if (r) {\n    return olen;\n  } else {\n    return -1;\n  }\n}")
  (c-define-type EVP_CIPHER "EVP_CIPHER")
  (c-define-type EVP_CIPHER* (pointer EVP_CIPHER (EVP_CIPHER*)))
  (c-define-type EVP_CIPHER_CTX "EVP_CIPHER_CTX")
  (c-define-type
   EVP_CIPHER_CTX*
   (pointer EVP_CIPHER_CTX (EVP_CIPHER_CTX*) "ffi_release_EVP_CIPHER_CTX"))
  (define-c-type-predicate EVP_CIPHER? EVP_CIPHER*)
  (define-c-type-predicate EVP_CIPHER_CTX? EVP_CIPHER_CTX*)
  (define-c-lambda
   EVP_CIPHER_CTX_create
   ()
   EVP_CIPHER_CTX*
   "ffi_create_EVP_CIPHER_CTX")
  (define-c-lambda
   EVP_EncryptInit
   (EVP_CIPHER_CTX* EVP_CIPHER* scheme-object scheme-object)
   int
   "ffi_EVP_EncryptInit")
  (define-c-lambda
   EVP_EncryptUpdate
   (EVP_CIPHER_CTX* scheme-object int scheme-object int int)
   int
   "ffi_EVP_EncryptUpdate")
  (define-c-lambda
   EVP_EncryptFinal
   (EVP_CIPHER_CTX* scheme-object int)
   int
   "ffi_EVP_EncryptFinal")
  (define-c-lambda
   EVP_DecryptInit
   (EVP_CIPHER_CTX* EVP_CIPHER* scheme-object scheme-object)
   int
   "ffi_EVP_DecryptInit")
  (define-c-lambda
   EVP_DecryptUpdate
   (EVP_CIPHER_CTX* scheme-object int scheme-object int int)
   int
   "ffi_EVP_DecryptUpdate")
  (define-c-lambda
   EVP_DecryptFinal
   (EVP_CIPHER_CTX* scheme-object int)
   int
   "ffi_EVP_DecryptFinal")
  (define-c-lambda
   EVP_CIPHER_CTX_copy
   (EVP_CIPHER_CTX* EVP_CIPHER_CTX*)
   int
   "EVP_CIPHER_CTX_copy")
  (define-c-lambda/const-pointer EVP_rc4 () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_aes_128_ecb () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_aes_128_cbc () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_aes_128_cfb () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_aes_128_ofb () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_aes_128_ctr () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_aes_128_ccm () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_aes_128_gcm () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_aes_128_xts () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_aes_192_ecb () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_aes_192_cbc () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_aes_192_cfb () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_aes_192_ofb () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_aes_192_ctr () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_aes_192_ccm () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_aes_192_gcm () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_aes_256_ecb () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_aes_256_cbc () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_aes_256_cfb () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_aes_256_ofb () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_aes_256_ctr () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_aes_256_ccm () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_aes_256_gcm () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_aes_256_xts () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_camellia_128_ecb () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_camellia_128_cbc () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_camellia_128_cfb () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_camellia_128_ofb () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_camellia_192_ecb () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_camellia_192_cbc () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_camellia_192_cfb () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_camellia_192_ofb () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_camellia_256_ecb () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_camellia_256_cbc () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_camellia_256_cfb () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_camellia_256_ofb () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_cast5_ecb () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_cast5_cbc () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_cast5_cfb () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_cast5_ofb () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_bf_ecb () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_bf_cbc () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_bf_cfb () EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_bf_ofb () EVP_CIPHER*)
  (define-c-lambda EVP_CIPHER_nid (EVP_CIPHER*) int)
  (define-c-lambda EVP_CIPHER_block_size (EVP_CIPHER*) int)
  (define-c-lambda EVP_CIPHER_key_length (EVP_CIPHER*) int)
  (define-c-lambda EVP_CIPHER_iv_length (EVP_CIPHER*) int)
  (define-c-lambda/const-pointer EVP_CIPHER_name (EVP_CIPHER*) char-string)
  (define-c-lambda/const-pointer
   EVP_CIPHER_CTX_cipher
   (EVP_CIPHER_CTX*)
   EVP_CIPHER*)
  (define-c-lambda EVP_CIPHER_CTX_nid (EVP_CIPHER_CTX*) int)
  (define-c-lambda EVP_CIPHER_CTX_block_size (EVP_CIPHER_CTX*) int)
  (define-c-lambda EVP_CIPHER_CTX_key_length (EVP_CIPHER_CTX*) int)
  (define-c-lambda EVP_CIPHER_CTX_iv_length (EVP_CIPHER_CTX*) int)
  (define-c-lambda/const-pointer
   EVP_get_cipherbyname
   (char-string)
   EVP_CIPHER*)
  (define-c-lambda/const-pointer EVP_get_cipherbynid (int) EVP_CIPHER*)
  (c-declare
   "static HMAC_CTX *ffi_create_HMAC_CTX ()\n{\n#if OPENSSL_VERSION_NUMBER < 0x10100000L ||  defined (LIBRESSL_VERSION_NUMBER)\n  HMAC_CTX *ctx = (HMAC_CTX*)malloc (sizeof (HMAC_CTX));\n  if (ctx) {\n    HMAC_CTX_init (ctx);\n  }\n#else\n  HMAC_CTX *ctx = HMAC_CTX_new ();\n#endif\n  return ctx;\n}\n\nstatic ___SCMOBJ ffi_release_HMAC_CTX (void *ptr)\n{\n#if OPENSSL_VERSION_NUMBER < 0x10100000L ||  defined (LIBRESSL_VERSION_NUMBER)\n  HMAC_CTX_cleanup ((HMAC_CTX*) ptr);\n  free (ptr);\n#else\n  HMAC_CTX_free ((HMAC_CTX*) ptr);\n#endif\n  return ___FIX (___NO_ERR);\n}\n\nstatic int ffi_HMAC_Init (HMAC_CTX *ctx, ___SCMOBJ key, EVP_MD *type)\n{\n  return HMAC_Init_ex (ctx, U8_DATA (key), U8_LEN (key), type, NULL);\n}\n\nstatic int ffi_HMAC_Update (HMAC_CTX *ctx, ___SCMOBJ bytes, int start, int end)\n{\n  return HMAC_Update (ctx, U8_DATA (bytes) + start, end - start);\n}\n\nstatic int ffi_HMAC_Final (HMAC_CTX *ctx, ___SCMOBJ bytes)\n{\n  return HMAC_Final (ctx, U8_DATA (bytes), NULL);\n}\n\nstatic void ffi_HMAC (EVP_MD *type, ___SCMOBJ key, ___SCMOBJ bytes, int start, int end, ___SCMOBJ digest)\n{\n  HMAC (type, U8_DATA (key), U8_LEN (key), U8_DATA (bytes) + start, end - start, U8_DATA (digest), NULL);\n}")
  (c-define-type HMAC_CTX "HMAC_CTX")
  (c-define-type
   HMAC_CTX*
   (pointer HMAC_CTX (HMAC_CTX*) "ffi_release_HMAC_CTX"))
  (define-c-type-predicate HMAC_CTX? HMAC_CTX*)
  (define-c-lambda HMAC_CTX_create () HMAC_CTX* "ffi_create_HMAC_CTX")
  (define-c-lambda
   HMAC_Init
   (HMAC_CTX* scheme-object EVP_MD*)
   int
   "ffi_HMAC_Init")
  (define-c-lambda
   HMAC_Update
   (HMAC_CTX* scheme-object int int)
   int
   "ffi_HMAC_Update")
  (define-c-lambda HMAC_Final (HMAC_CTX* scheme-object) int "ffi_HMAC_Final")
  (define-c-lambda
   HMAC
   (EVP_MD* scheme-object scheme-object int int scheme-object)
   void
   "ffi_HMAC")
  (c-declare
   "static ___SCMOBJ ffi_BN_free (void *bn)\n{\n BN_free ((BIGNUM *)bn);\n return ___FIX (___NO_ERR);\n}\n\nstatic BIGNUM *ffi_BN_bin2bn (___SCMOBJ data, int start, int end)\n{\n  return BN_bin2bn (U8_DATA (data) + start, end -start, NULL);\n}\n\nstatic int ffi_BN_bn2bin (BIGNUM *bn, ___SCMOBJ data)\n{\n return BN_bn2bin (bn, U8_DATA (data));\n}")
  (c-define-type BN "BIGNUM")
  (c-define-type BN* (pointer BN (BN*) "ffi_BN_free"))
  (define-c-type-predicate BN? BN*)
  (define-c-lambda BN_num_bytes (BN*) int)
  (define-c-lambda BN_bin2bn (scheme-object int int) BN* "ffi_BN_bin2bn")
  (define-c-lambda BN_bn2bin (BN* scheme-object) int "ffi_BN_bn2bin")
  (c-declare
   "static ___SCMOBJ ffi_DH_free (void *dh)\n{\n DH_free ((DH *)dh);\n return ___FIX (___NO_ERR);\n}\n\nstatic BIGNUM *ffi_DH_pub_key (DH *dh)\n{\n#if OPENSSL_VERSION_NUMBER < 0x10100000L ||  defined (LIBRESSL_VERSION_NUMBER)\n return dh->pub_key;\n#else\n BIGNUM const *pub;\n DH_get0_key (dh, &pub, NULL);\n return (BIGNUM*) pub;\n#endif\n}\n\nstatic int ffi_DH_compute_key (___SCMOBJ secret, BIGNUM *pubkey, DH *dh)\n{\n return DH_compute_key (U8_DATA (secret), pubkey, dh);\n}\n\n#if OPENSSL_VERSION_NUMBER < 0x10002000L || defined (LIBRESSL_VERSION_NUMBER)\n#include \"libcrypto-rfc5114.c\"\n#endif")
  (c-define-type BN-borrow* (pointer BN (BN*)))
  (c-define-type DH "DH")
  (c-define-type DH* (pointer DH (DH*) "ffi_DH_free"))
  (define-c-type-predicate DH? DH*)
  (define-c-lambda DH_pub_key (DH*) BN-borrow* "ffi_DH_pub_key")
  (define-c-lambda DH_new () DH*)
  (define-c-lambda DH_get_1024_160 () DH*)
  (define-c-lambda DH_get_2048_224 () DH*)
  (define-c-lambda DH_get_2048_256 () DH*)
  (define-c-lambda DH_generate_key (DH*) int)
  (define-c-lambda DH_size (DH*) int)
  (define-c-lambda
   DH_compute_key
   (scheme-object BN* DH*)
   int
   "ffi_DH_compute_key")
  (c-declare
   "static ___SCMOBJ ffi_release_EVP_PKEY (void *ptr)\n{\n  EVP_PKEY_free ((EVP_PKEY*)ptr);\n  return ___FIX (___NO_ERR);\n}\nstatic ___SCMOBJ ffi_release_EVP_PKEY_CTX (void *ptr)\n{\n  EVP_PKEY_CTX_free ((EVP_PKEY_CTX*)ptr);\n  return ___FIX (___NO_ERR);\n}\nstatic EVP_PKEY* ffi_EVP_PKEY_keygen (EVP_PKEY_CTX* ctx) {\n  EVP_PKEY* pkey = NULL; // NB: not initializing this variable leads to occasional segfaults\n  if (EVP_PKEY_keygen(ctx, &pkey) == 1) {\n    return pkey;\n  } else {\n    return NULL;\n  }\n}\nstatic EVP_PKEY *ffi_EVP_PKEY_new_raw_private_key (int type, ENGINE* e, ___SCMOBJ o)\n{\n#if OPENSSL_VERSION_NUMBER >= 0x10101000L\n  return EVP_PKEY_new_raw_private_key(type, e, U8_DATA(o), (size_t)U8_LEN(o));\n#else\n  return NULL;\n#endif\n}\nstatic EVP_PKEY *ffi_EVP_PKEY_new_raw_public_key (int type, ENGINE* e, ___SCMOBJ o)\n{\n#if OPENSSL_VERSION_NUMBER >= 0x10101000L\n  return EVP_PKEY_new_raw_public_key(type, e, U8_DATA(o), (size_t)U8_LEN(o));\n#else\n  return NULL;\n#endif\n}\nstatic int ffi_EVP_PKEY_get_raw_private_key (EVP_PKEY* pkey, ___SCMOBJ o)\n{\n#if OPENSSL_VERSION_NUMBER >= 0x10101000L\n  size_t len;\n  if ((void*)o == (void*)___FAL) {\n    EVP_PKEY_get_raw_private_key(pkey, NULL, &len);\n    return (int)len;\n  } else if (EVP_PKEY_get_raw_private_key(pkey, U8_DATA(o), &len)) {\n    return (int)len;\n  } else {\n    return 0;\n  }\n#else\n  return 0;\n#endif\n}\nstatic int ffi_EVP_PKEY_get_raw_public_key (EVP_PKEY* pkey, ___SCMOBJ o)\n{\n#if OPENSSL_VERSION_NUMBER >= 0x10101000L\n  size_t len;\n  if ((void*)o == (void*)___FAL) {\n    EVP_PKEY_get_raw_public_key(pkey, NULL, &len);\n    return (int)len;\n  } else {\n    return EVP_PKEY_get_raw_public_key(pkey, U8_DATA(o), &len) ? (int)len : 0;\n  }\n#else\n  return 0;\n#endif\n}\nstatic int ffi_EVP_DigestSign(EVP_MD_CTX *ctx, ___SCMOBJ sig, ___SCMOBJ tbs)\n{\n  size_t siglen = U8_LEN(sig);\n#if OPENSSL_VERSION_NUMBER >= 0x10101000L\n  return EVP_DigestSign(ctx, U8_DATA(sig), &siglen, U8_DATA(tbs), U8_LEN(tbs)) ? siglen : 0;\n#else\n  return 0;\n#endif\n}\nstatic int ffi_EVP_DigestVerify(EVP_MD_CTX *ctx, ___SCMOBJ sig, ___SCMOBJ tbs)\n{\n#if OPENSSL_VERSION_NUMBER >= 0x10101000L\n  return EVP_DigestVerify(ctx, U8_DATA(sig), U8_LEN(sig), U8_DATA(tbs), U8_LEN(tbs));\n#else\n  return 0;\n#endif\n}\n\n#if OPENSSL_VERSION_NUMBER < 0x10101000L\nvoid EVP_MD_CTX_set_pkey_ctx(EVP_MD_CTX *ctx, EVP_PKEY_CTX *pctx) { return; }\n#endif")
  (c-define-type EVP_PKEY "EVP_PKEY")
  (c-define-type EVP_PKEY* (pointer EVP_PKEY (EVP_PKEY*)))
  (c-define-type EVP_PKEY_CTX "EVP_PKEY_CTX")
  (c-define-type
   EVP_PKEY_CTX*
   (pointer EVP_PKEY_CTX (EVP_PKEY_CTX*) "ffi_release_EVP_PKEY_CTX"))
  (define-macro (define-consts . cs)
    `(begin ,@(map (lambda (c) `(define-const ,c)) cs)))
  (define-macro (define-consts* ccond . cs)
    `(begin ,@(map (lambda (c) `(define-const* ,c ,ccond)) cs)))
  (define-consts
   EVP_PKEY_NONE
   EVP_PKEY_RSA
   EVP_PKEY_RSA2
   EVP_PKEY_DSA
   EVP_PKEY_DSA1
   EVP_PKEY_DSA2
   EVP_PKEY_DSA3
   EVP_PKEY_DSA4
   EVP_PKEY_DH
   EVP_PKEY_EC
   EVP_PKEY_HMAC)
  (define-consts*
   "OPENSSL_VERSION_NUMBER >= 0x10100000L"
   EVP_PKEY_DHX
   EVP_PKEY_CMAC
   EVP_PKEY_TLS1_PRF
   EVP_PKEY_HKDF)
  (define-consts*
   "OPENSSL_VERSION_NUMBER >= 0x10101000L"
   EVP_PKEY_RSA_PSS
   EVP_PKEY_SM2
   EVP_PKEY_SCRYPT
   EVP_PKEY_SIPHASH
   EVP_PKEY_POLY1305
   EVP_PKEY_X25519
   EVP_PKEY_ED25519
   EVP_PKEY_X448
   EVP_PKEY_ED448)
  (define-c-lambda EVP_PKEY_CTX_new (EVP_PKEY* ENGINE*) EVP_PKEY_CTX*)
  (define-c-lambda EVP_PKEY_CTX_new_id (int ENGINE*) EVP_PKEY_CTX*)
  (define-c-lambda EVP_PKEY_CTX_dup (EVP_PKEY_CTX*) EVP_PKEY_CTX*)
  (define-c-lambda EVP_PKEY_keygen_init (EVP_PKEY_CTX*) int)
  (define-c-lambda
   EVP_PKEY_keygen
   (EVP_PKEY_CTX*)
   EVP_PKEY*
   "ffi_EVP_PKEY_keygen")
  (define-c-lambda
   EVP_PKEY_new_raw_private_key
   (int ENGINE* scheme-object)
   EVP_PKEY*
   "ffi_EVP_PKEY_new_raw_private_key")
  (define-c-lambda
   EVP_PKEY_new_raw_public_key
   (int ENGINE* scheme-object)
   EVP_PKEY*
   "ffi_EVP_PKEY_new_raw_public_key")
  (define-c-lambda
   EVP_PKEY_get_raw_private_key
   (EVP_PKEY* scheme-object)
   int
   "ffi_EVP_PKEY_get_raw_private_key")
  (define-c-lambda
   EVP_PKEY_get_raw_public_key
   (EVP_PKEY* scheme-object)
   int
   "ffi_EVP_PKEY_get_raw_public_key")
  (define-c-lambda
   EVP_DigestSign
   (EVP_MD_CTX* scheme-object scheme-object)
   int
   "ffi_EVP_DigestSign")
  (define-c-lambda
   EVP_DigestVerify
   (EVP_MD_CTX* scheme-object scheme-object)
   int
   "ffi_EVP_DigestVerify")
  (define-c-lambda EVP_MD_CTX_set_pkey_ctx (EVP_MD_CTX* EVP_PKEY_CTX*) void)
  (c-declare
   "#ifndef ___HAVE_FFI_FREE\n#define ___HAVE_FFI_FREE\n___SCMOBJ ffi_free (void *ptr)\n{\n free (ptr);\n return ___FIX (___NO_ERR);\n}\n#endif"))
